//
// amsgen.cpp
//
// This software was written by Leonard Janke (janke@unixg.ubc.ca)
// in 1996-7 and is entered, by him, into the public domain.

#include <iomanip.h>
#include <stdlib.h>
#include "asmgen.hpp"

// *****************************************************

// these defaults don't really matter since they can be change at
// runtime

#define DEFAULTSYNTAX MASM
// #define DEFSYNTAX NASM 
// #define DEFSYNTAX ATT

#define DEFAULTTARGET WIN32
// #define DEFAULTTARGET AOUT
// #define DEFAULTTARGET COFF
// #define DEFAULTTARGET CPP

#define DEFAULTCALLINGCONVENTION CDECL
// #define DEFAULTCALLINGCONVENTION=WATCOMREGISTER

#define DEFAULTCPU 386
// #define DEFAULTCPU 486
// #define DEFAULTCPU 586

// Some versions of GNU's as and MSVC++ (include version 4.0) do not
// encode rol and ror by 1 properly. Leave this macro defined to work
// around this bug, or leave it defined just to be safe.
//
// Thanks to Peter Gutmann for advice on how to handle this bug. -Leonard

#define ROTATEBY1BUG

// *****************************************************

const Register eax(Register::eax);
const Register ebx(Register::ebx);
const Register ecx(Register::ecx);
const Register edx(Register::edx);
const Register edi(Register::edi);
const Register esi(Register::esi);
const Register ebp(Register::ebp);
const Register esp(Register::esp);

AsmGenerator::syntaxEnum 
AsmGenerator::syntax=AsmGenerator::DEFAULTSYNTAX;

AsmGenerator::targetEnum 
AsmGenerator::target=AsmGenerator::DEFAULTTARGET;

AsmGenerator::callingConventionEnum 
AsmGenerator::callingConvention=AsmGenerator::DEFAULTCALLINGCONVENTION;

int AsmGenerator::TheCPU=DEFAULTCPU;

const char* AsmGenerator::IntelName[8]=
{
  "eax", "ebx", "ecx", "edx", "edi", "esi", "ebp", "esp"
};

const char* AsmGenerator::ATTName[8]=
{
  "%eax", "%ebx", "%ecx", "%edx", "%edi", "%esi", "%ebp", "%esp"
};

int AsmGenerator::RegCode(const Register& which)
{
  int rv;
  switch ( which._register )
    {
    case Register::eax:
      rv=0;
      break;
    case Register::ebx:
      rv=3;
      break;
    case Register::ecx:
      rv=1;
      break;
    case Register::edx:
      rv=2;
      break;
    case Register::edi:
      rv=7;
      break;
    case Register::esi:
      rv=6;
      break;
    case Register::ebp:
      rv=5;
      break;
    case Register::esp:
      rv=4;
      break;
    default:
      cerr <<"Ain't no register I ever heard about before."<<endl;
      exit(-1);
    }

  return rv;
}

Register::operator const char*() const
{
  const char* registerName;
  switch ( AsmGenerator::syntax )
    {
    case AsmGenerator::MASM:
    case AsmGenerator::NASM:
      registerName=AsmGenerator::IntelName[_register];
      break;
    case AsmGenerator::ATT:
      registerName=AsmGenerator::ATTName[_register];
      break;
    default:
      cerr <<"Unrecognized syntax"<<endl;
      exit(-1);
    }

   return registerName;
}

AsmGenerator::AsmGenerator(ostream& os) : os(os)
{
}

void AsmGenerator::Comment(const char* comment)
{
  switch ( syntax )
    {
    case NASM:
      os <<"; "<<comment<<endl;
      break;
    case MASM:
      switch ( target )
        {
        case WIN32:
          os <<"; "<<comment<<endl;
          break;
        case INL:
          os <<"/* "<<comment<<" */"<<endl;
          break;
        default:
          cerr <<"Target not supported for this syntax"<<endl;
          exit(-1);
        }
      break;
    case ATT:
      os <<"# "<<comment<<endl;
      break;
    default:
      cerr <<"Unrecognized syntax"<<endl;
      exit(-1);
    }
}

void AsmGenerator::WriteAsmFunctionName()
{
  switch ( target )
    {
    case ELF:
      os <<_cFunctionName;
      break;
    case AOUT:
    case COFF:
    case WIN32:
      switch ( callingConvention )
        {
        case CDECL:
	  os <<"_"<<_cFunctionName;
	  break;
        case WATCOMREGISTER:
	  os <<_cFunctionName<<"_";
	  break;
        default:
          cerr <<"Unrecognized calling convention"<<endl;
          exit(-1);
        }
      break;
    case CPP:
      // Function name will be generated by the C preprocessor using
      // some macros defined in asm.h due to Eric Young and sent to me by  
      // Peter Gutmann. Thanks to Peter Gutmann for suggesting
      // this.  - Leonard
      os <<"FUNCTION("<<_cFunctionName<<")";
      break;
    }
}

void AsmGenerator::FileBegin()
{
  switch ( syntax )
    {
    case MASM:
      os <<"."<<TheCPU<<endl;
      os <<".MODEL FLAT"<<endl;
      os <<"_TEXT SEGMENT"<<endl;
      os <<"PUBLIC "; WriteAsmFunctionName(); os <<endl;
      WriteAsmFunctionName(); os <<" PROC"<<endl;
      break;
    case NASM:
      os <<"[BITS 32]"<<endl;
      os <<"[SECTION .text]"<<endl;
      switch ( target )
	{
	case ELF:
	case AOUT:
	case COFF:
	case WIN32:
	  os <<"[GLOBAL "; WriteAsmFunctionName(); os <<"]"<<endl;
	  WriteAsmFunctionName(); os <<":"<<endl;
	  break;
	default:
	  cerr <<"Unrecognized target"<<endl;
	  exit(-1);
	}
      break;
    case ATT:
      switch ( target )
	{
	case ELF:
	case AOUT:
	case COFF:
        case WIN32: 
	  os <<".text"<<endl;
	  os <<".globl "; WriteAsmFunctionName(); os <<endl;
	  os <<".align "<<((target==ELF) ? 16 : 4)<<endl;
	  WriteAsmFunctionName(); os <<":"<<endl;
	  break;
	case CPP:  
          // Use the C preprocessor to help generate the function beginning,
          // using some macros defined in asm.h due to Eric Young and 
          // sent to me by  Peter Gutmann. Thanks to Peter Gutmann for 
          // suggesting this.  - Leonard

	  os <<"#include \"asm.h\""<<endl;
	  os <<".text"<<endl;
	  os <<".globl FUNCTION("<<_cFunctionName<<")"<<endl;
	  os <<".align ALIGN"<<endl;
	  os <<"FUNCTION("<<_cFunctionName<<")"<<endl;
	  break;
	default:
	  cerr <<"Unrecognized target"<<endl;
	  exit(-1);
	}
      break;
    default:
      cerr <<"Unrecognized syntax"<<endl;
      exit(-1);
    }
}

void AsmGenerator::Subround(const int number)
{
  switch ( syntax )
    {
    case NASM:
      os <<"; Subround "<<number<<endl;
      break;
    case MASM:
      switch ( target )
        {
        case WIN32:
          os <<"; Subround "<<number<<endl;
          break;
        case INL:
          os <<"/* Subround "<<number<<" */"<<endl;
          break;
        default:
          cerr <<"Target not supported for this syntax"<<endl;
          exit(-1);
        }
      break;
    case ATT:
      os <<"# Subround "<<number<<endl;
      break;
    default:
      cerr <<"Unrecognized syntax"<<endl;
      exit(-1);
    }
}

void AsmGenerator::PushOffsetAddress(const Register& base, const int offset)
{
  switch ( syntax )
    {
    case MASM:
    case NASM:
      if ( offset )
	os <<"["<<base<<"+"<<offset<<"]";
      else
	os <<"["<<base<<"]";
      break;
    case ATT:
      if (offset)
	os <<offset<<"("<<base<<")";
      else
	os <<"("<<base<<")";
      break;
    default:
      cerr <<"Unrecognized syntax"<<endl;
      exit(-1);
    }
}

void AsmGenerator::PushIndexedAddress(const Register& base, const int scale, 
				      const Register& index, const int offset)
{
  switch ( syntax )
    {
    case MASM:
    case NASM:
      os <<"["<<base;
      if ( scale )
	if ( scale != 1 )
	  os <<"+"<<scale<<"*"<<index;
	else
	  os <<"+"<<index;
      if (offset)
	os <<"+"<<offset;
      os <<"]";
      break;
    case ATT:
      if (offset)
	os <<offset;
      os <<"("<<base;
      if ( scale )
	os <<","<<index<<","<<scale;
      os <<")";
      break;
    default:
      cerr <<"Unrecognized syntax"<<endl;
      exit(-1);
    }
}
 
void AsmGenerator::Mov(const Register& dest, const Register& src)
{
  switch ( syntax )
    {
    case MASM:
    case NASM:
      os <<"\tmov "<<dest<<", "<<src<<endl;
      break;
    case ATT:
      os <<"\tmovl "<<src<<", "<<dest<<endl;
      break;
    default:
      cerr <<"Unrecognized syntax"<<endl;
      exit(-1);
    }
}

void AsmGenerator::Mov(const Register& dest, const s32 immediate)
{
  switch ( syntax )
    {
    case MASM:
    case NASM:
      os <<"\tmov "<<dest<<", "<<immediate<<endl;
      break;
    case ATT:
      os <<"\tmovl $"<<immediate<<", "<<dest<<endl;
      break;
    default:
      cerr <<"Unrecognized syntax"<<endl;
      exit(-1);
    }
}
 
void AsmGenerator::Mov(const Register& dest, const Register& src,
		       const int offset)
{
  switch ( syntax )
    {
    case MASM:
    case NASM:
      os <<"\tmov "<<dest<<", "; PushOffsetAddress(src,offset); os <<endl;
      break;
    case ATT:
      os <<"\tmovl "; PushOffsetAddress(src,offset); os <<", "<<dest<<endl;
      break;
    default:
      cerr <<"Unrecognized syntax"<<endl;
      exit(-1);
    }
}
 
void AsmGenerator::Mov(const Register& dest, const int offset, 
		       const Register& src)
{
  switch ( syntax )
    {
    case MASM:
    case NASM:
      os <<"\tmov "; PushOffsetAddress(dest,offset); os <<", "<<src<<endl;
      break;
    case ATT:
      os <<"\tmovl "<<src<<", "; PushOffsetAddress(dest,offset); os <<endl;
      break;
    default:
      cerr <<"Unrecognized syntax"<<endl;
      exit(-1);
    }
}
 
void AsmGenerator::And(const Register& dest, const Register& src)
{
  switch ( syntax )
    {
    case MASM:
    case NASM:
      os <<"\tand "<<dest<<", "<<src<<endl;
      break;
    case ATT:
      os <<"\tandl "<<src<<", "<<dest<<endl;
      break;
    default:
      cerr <<"Unrecognized syntax"<<endl;
      exit(-1);
    }
}
 
void AsmGenerator::Or(const Register& dest, const Register& src)
{
  switch ( syntax )
    {
    case MASM:
    case NASM:
      os <<"\tor "<<dest<<", "<<src<<endl;
      break;
    case ATT:
      os <<"\torl "<<src<<", "<<dest<<endl;
      break;
    default:
      cerr <<"Unrecognized syntax"<<endl;
      exit(-1);
    }
}
 
void AsmGenerator::Xor(const Register& dest, const Register& src)
{
  switch ( syntax )
    {
    case MASM:
    case NASM:
      os <<"\txor "<<dest<<", "<<src<<endl;
      break;
    case ATT:
      os <<"\txorl "<<src<<", "<<dest<<endl;
      break;
    default:
      cerr <<"Unrecognized syntax"<<endl;
      exit(-1);
    }
}
 
void AsmGenerator::Xor(const Register& dest, const s32 immediate)
{
  switch ( syntax )
    {
    case MASM:
    case NASM:
      os <<"\txor "<<dest<<", "<<immediate<<endl;
      break;
    case ATT:
      os <<"\txorl $"<<immediate<<", "<<dest<<endl;
      break;
    default:
      cerr <<"Unrecognized syntax"<<endl;
      exit(-1);
    }
}
 
void AsmGenerator::Add(const Register& dest, const Register& src)
{
  switch ( syntax )
    {
    case MASM:
    case NASM:
      os <<"\tadd "<<dest<<", "<<src<<endl;
      break;
    case ATT:
      os <<"\taddl "<<src<<", "<<dest<<endl;
      break;
    default:
      cerr <<"Unrecognized syntax"<<endl;
      exit(-1);
    }
}

void AsmGenerator::Add(const Register& dest, const Register& src, 
		       const int offset)
{
  switch ( syntax )
    {
    case MASM:
    case NASM:
      os <<"\tadd "<<dest<<", "; PushOffsetAddress(src,offset); os <<endl;
      break;
    case ATT:
      os <<"\taddl "; PushOffsetAddress(src,offset); os <<", "<<dest<<endl;
      break;
    default:
      cerr <<"Unrecognized syntax"<<endl;
      exit(-1);
    }
}
 
void AsmGenerator::Add(const Register& dest, const s32 immediate)
{
  switch ( syntax )
    {
    case MASM:
    case NASM:
      os <<"\tadd "<<dest<<", "<<long(immediate)<<endl;
      break;
    case ATT:
      os <<"\taddl $"<<long(immediate)<<", "<<dest<<endl;
      break;
    default:
      cerr <<"Unrecognized syntax"<<endl;
      exit(-1);
    }
}

void AsmGenerator::Sub(const Register& dest, const s32 immediate)
{
  switch ( syntax )
    {
    case MASM:
    case NASM:
      os <<"\tsub "<<dest<<", "<<long(immediate)<<endl;
      break;
    case ATT:
      os <<"\tsubl $"<<long(immediate)<<", "<<dest<<endl;
      break;
    default:
      cerr <<"Unrecognized syntax"<<endl;
      exit(-1);
    }
}
 
void AsmGenerator::Rol(const Register& dest, const int distance)
{
  switch ( syntax )
    {
    case MASM:
#if defined(ROTATEBY1BUG)
    if ( target != INL || distance!=1 )
      os <<"\trol "<<dest<<", "<<distance<<endl;
    else
    {
      os <<"\t_emit 0xd1"<<endl;
      os <<"\t_emit 0xc"<<RegCode(dest)<<endl;
    }
#else
      os <<"\trol "<<dest<<", "<<distance<<endl;
#endif
     break;
    case NASM:
      os <<"\trol "<<dest<<", "<<distance<<endl;
      break;
    case ATT:

#if defined(ROTATEBY1BUG)
      if ( distance!=1 )
	os <<"\troll $"<<distance<<", "<<dest<<endl;
      else
	os <<"\t.byte 0xd1, 0xc"<<RegCode(dest)<<endl;
#else
      os <<"\troll $"<<distance<<", "<<dest<<endl;
#endif
      break;
    default:
      cerr <<"Unrecognized syntax"<<endl;
      exit(-1);
    }
}

void AsmGenerator::Ror(const Register& dest, const int distance)
{
  switch ( syntax )
    {
    case MASM:
#if defined(ROTATEBY1BUG)
      if ( target != INL || distance!=1 )
        os <<"\tror "<<dest<<", "<<distance<<endl;
      else
      {
        os <<"\t_emit 0xd1"<<endl;
        os <<"\t_emit "<<(12*16+8+RegCode(dest))<<endl;
      }
#else
        os <<"\tror "<<dest<<", "<<distance<<endl;
#endif
      break;
    case NASM:
      os <<"\tror "<<dest<<", "<<distance<<endl;
      break;
    case ATT:
#if defined(ROTATEBY1BUG)
      if ( distance!=1 )
	os <<"\trorl $"<<distance<<", "<<dest<<endl;
      else
	os <<"\t.byte 0xd1, "<<(12*16+8+RegCode(dest))<<endl;
#else
      os <<"\trorl $"<<distance<<", "<<dest<<endl;
#endif
      break;
    default:
      cerr <<"Unrecognized syntax"<<endl;
      exit(-1);
    }
}

void AsmGenerator::BSwap(const Register& dest)
{
  switch ( syntax )
    {
    case MASM:
    case NASM:
      os <<"\tbswap "<<dest<<endl;
      break;
    case ATT:
      os <<"\tbswapl "<<dest<<endl;
      break;
    default:
      cerr <<"Unrecognized syntax"<<endl;
      exit(-1);
    }
}
 
void AsmGenerator::Push(const Register& src)
{
  switch ( syntax )
    {
    case MASM:
    case NASM:
      os <<"\tpush "<<src<<endl;
      break;
    case ATT:
      os <<"\tpushl "<<src<<endl;
      break;
    default:
      cerr <<"Unrecognized syntax"<<endl;
      exit(-1);
    }
}
 
void AsmGenerator::Pop(const Register& dest)
{
  switch ( syntax )
    {
    case MASM:
    case NASM:
      os <<"\tpop "<<dest<<endl;
      break;
    case ATT:
      os <<"\tpopl "<<dest<<endl;
      break;
    default:
      cerr <<"Unrecognized syntax"<<endl;
      exit(-1);
    }
}

void AsmGenerator::Nop()
{
  switch ( syntax )
    {
    case MASM:
    case NASM:
    case ATT:
      os <<"\tnop"<<endl;
      break;
    default:
      cerr <<"Unrecognized syntax"<<endl;
      exit(-1);
    }
}

void AsmGenerator::Ret(const int adjustment)
{
  os <<"\tret "<<adjustment<<endl;
}
 
void AsmGenerator::Ret()
{
  switch ( syntax )
    {
    case MASM:
    case NASM:
      os <<"\tret"<<endl;
      break;
    case ATT:
      os <<"\tret"<<endl;
      break;
    default:
      cerr <<"Unrecognized syntax"<<endl;
      exit(-1);
    }
}
 
void AsmGenerator::Lea(const Register& dest, const Register& src, 
		       const int scale, const Register& index, 
		       const int offset)
{
  switch ( syntax )
    {
    case MASM:
    case NASM:
      os <<"\tlea "<<dest<<", "; 
      PushIndexedAddress(src,scale,index,offset);
      os <<endl;
      break;
    case ATT:
      os <<"\tleal ";
      PushIndexedAddress(src,scale,index,offset);
      os <<", "<<dest<<endl;
      break;
    default:
      cerr <<"Unrecognized syntax"<<endl;
      exit(-1);
    }
}

void AsmGenerator::FileEnd()
{
  switch ( syntax )
    {
    case NASM:
    case ATT:
      break;
    case MASM:
      WriteAsmFunctionName(); os <<" ENDP"; cout <<endl;
      os <<"_TEXT ENDS"<<endl;
      os <<"END"<<endl;
      break;
    default:
      cerr <<"Unrecognized syntax"<<endl;
      exit(-1);
    }
}
